#=======================================================================
#	$Id: WebFiler.pm,v 1.3 2009/06/26 10:30:57 pythontech Exp $
#	Serve files via http (simpler and less restrictive than DAV).
#=======================================================================
package PythonTech::WebFiler;
use strict;
use HTTP::Date;
use PythonTech::Conv qw(uq hq);

#-----------------------------------------------------------------------
#	Create new server object
#	Configuration options:
#	  basePath		Root of filestore to be exported
#	  authentication	How to authenticate remote user:
#		none		Allow anybody access (unsafe!)
#		htaccess	Assume script is protected by .htaccess
#				rules e.g. "require valid-user" or
#				"require group foobar"
#		cookie		User is associated with session cookie
#				using PythonTech::WebServer framework
#	  readPermission	Who has read permission.  Can be a Perl
#				list of usernames, or '*'
#	  writePermission	Who has write permission.  Can be a Perl
#				list of usernames, or '*'
#-----------------------------------------------------------------------
sub new {
    my $class = shift;
    my $self = {@_};
    foreach (qw(basePath)) {
	die "$class: $_ not configured\n" unless $self->{$_};
    }
    bless $self, $class;
    return $self;
}

sub webquery {
    my($self, $query) = @_;
    my $user;
    my $authtype = $self->{'authentication'};
    if ($authtype eq 'htaccess') {
	# Assume protected by .htaccess rule
	$user = $query->remote_user;
    } elsif ($authtype eq 'none') {
	# Unsafe!
	$user = '';
    } elsif ($authtype eq 'cookie') {
	# Crude authorisation: authentication
	$user = $query->require_login
	    or return;
    } else {
	die "Unknown authentication type: $authtype\n";
    }
    my $response = $query->response;
    my $method = $query->request_method;

    my $tpath = $query->param('path'); # Tainted
    my($path) = $tpath =~ m!^([^\0]*)$!
	or die "Invalid NUL in path\n";
    die "'..' not allowed\n" if "/$tpath/" =~ m!/\.\./!;
    my $fullpath = $self->{'basePath'} . $path;

    if ($method eq 'GET') {
	$self->permit($query, $user, 'read')
	    or return;
	if (-d $fullpath) {
	    my $listing;
	    opendir(PTWDIR,$fullpath) ||
		die "Cannot open directory $fullpath for read: $!\n";
	    foreach my $leaf (readdir(PTWDIR)) {
		next if $leaf eq '.' || $leaf eq '..';
		my $filename = "$fullpath/$leaf";
		my @stat = stat($filename);
		$listing .= sprintf("%s %s %o %d %d %s\n",
				    -l $filename ? "l" :
				    -d $filename ? "d" :
				    -f $filename ? "f" : "?",
				    uq($leaf),
				    $stat[2], # Mode
				    $stat[7], # Size
				    $stat[9], # Modtime
				    -l $filename ? uq(readlink($filename)) : "");
	    }
	    closedir(PTWDIR);
	    $response->set_type("application/x-ptwb-dir");
	    $response->write($listing);
	} elsif (-f $fullpath) {
	    my $mtime = (stat $fullpath)[9];
	    open(PTWFILE,'<',$fullpath) || 
		die "Cannot open $fullpath for read: $!\n";
	    binmode(PTWFILE);
	    $response->set_type("application/octet-stream");
	    $response->set_header(last_modified=>&HTTP::Date::time2str($mtime));
	    my $data;
	    while (read(PTWFILE,$data,4096)) {
		$response->write($data);
	    }
	    $response->write(''); # Flush, in case zero length file
	    close(PTWFILE);
	} else {
	    if ($response->can('not_found')) {
		$response->not_found();
	    } else {
		# Backward compatibility
		my $q = $response->{'cgi'};	# FIXME naughty - need 404 method
		print $q->header(-status=>404);
		print "<title>Not found</title>$fullpath not found";
	    }
	}
    } elsif ($method eq 'POST') {
	$self->permit($query, $user, 'write')
	    or return;
	my $action = $query->param('action');
	if ($action eq 'write') {
	    # Should be PUT method perhaps?
	    (my $dir = $fullpath) =~ s![^/]+$!!;
	    if (-l $fullpath) {
		die "Is a symlink\n";
	    } elsif (-d $fullpath) {
		die "Is a directory\n";
	    } elsif (-e $fullpath) {
		die "Not a plain file\n"
		    unless -f $fullpath;
	    } elsif (! -d $dir) {
		die "No such directory $dir\n";
	    }
	    my $fh = $query->upload('content')
		or die "No content file uploaded\n";
	    my $newpath = "$fullpath.webfilernew";
	    open(PTWFILE,'>',$newpath) or
		die "Cannot open $newpath for write: $!\n";
	    binmode(PTWFILE);
	    my $buffer;
	    while (read($fh,$buffer,4096)) {
		print PTWFILE $buffer;
	    }
	    close(PTWFILE);
	    rename($newpath,$fullpath)
		or die "Failed to rename $newpath to $fullpath: $!\n";
	    # Update metadata
	    my $mtime = 0 + $query->param('mtime');
	    if ($mtime) {
		utime $mtime,$mtime, $fullpath;
	    }
	    my $mode = $query->param('mode');
	    if (defined $mode) {
		chmod oct($mode) & 07777, $fullpath;
	    }
	    $self->no_content($response);
	} elsif ($action eq 'mkdir') {
	    (my $parent = $fullpath) =~ s![^/]+$!!;
	    if (-l $fullpath) {
		die "Is a symlink\n";
	    } elsif (-d $fullpath) {
		die "Is a directory\n";
	    } elsif (-e $fullpath) {
		die "File exists\n";
	    } elsif (! -d $parent) {
		die "No such directory $parent\n";
	    }
	    mkdir($fullpath) or
		die "Failed to create directory: $!\n";
	    my $mode = $query->param('mode');
	    if (defined $mode) {
		chmod oct($mode) & 07777, $fullpath;
	    }
	    $self->no_content($response);
	} else {
	    die "Action $action unknown\n";
	}
    } else {
	die "Method $method not supported\n";
    }
}

#-----------------------------------------------------------------------
#	Check that user has correct permission
#-----------------------------------------------------------------------
sub permit {
    my($self, $query,$user,$action) = @_;
    my $permitted;
    my $list = $self->{"${action}Permission"};
    if (! defined($list)) {
	$permitted = undef;
    } elsif (ref($list) eq 'ARRAY') {
	$permitted = grep {$user eq $_} @$list;
    } elsif (ref($list) eq 'CODE') {
	$permitted = $list->($user);
    } elsif ($list eq '*') {
	$permitted = 1;
    } else {
	$permitted = undef;
    }
    if (! $permitted) {
	my $response = $query->response;
	if ($response->can('set_status')) {
	    $response->set_status(403); # Forbidden
	    $response->write("<title>Access denied</title>".
			     "You do not have $action access");
	} else {
	    # Backward compatibility
	    my $q = $response->{'cgi'};	# FIXME naughty - need 403 method
	    print $q->header(-status=>404);
	    print "<title>Not found</title>",hq($query->url)," not found";
	    print("<title>Access denied</title>",
		  "You do not have $action access");
	}
	return undef;
    } else {
	return 1;
    }
}

sub no_content {
    my($self, $response) = @_;
    if ($response->can('no_content')) {
	$response->no_content;
    } elsif ($response->can('set_status')) {
	$response->set_status('204 No Content');
	$response->write('');
    } else {
	# Backward compatibility
	my $q = $response->{'cgi'};	# FIXME naughty - need 404 method
	print $q->header(-status=>204);
    }
}

1;
